/*
 *  Copyright (c) xlightweb.org, 2008 - 2009. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;

import java.io.Closeable;
import java.io.Flushable;
import java.io.IOException;
import java.io.InputStream;
import java.nio.BufferOverflowException;
import java.nio.ByteBuffer;
import java.nio.channels.Channels;
import java.nio.channels.FileChannel;
import java.nio.channels.GatheringByteChannel;
import java.nio.channels.ReadableByteChannel;
import java.nio.channels.WritableByteChannel;
import java.util.List;

import org.xsocket.DataConverter;
import org.xsocket.IDataSink;
import org.xsocket.IDestroyable;
import org.xsocket.connection.IConnection;
import org.xsocket.connection.IWriteCompletionHandler;
import org.xsocket.connection.IConnection.FlushMode;





/**
 * I/O resource capable of receiving the body data.
 * 
 * @author grro@xlightweb.org
 *
 */
public abstract class BodyDataSink implements IDataSink, IDestroyable, Flushable, Closeable, WritableByteChannel, GatheringByteChannel {

    private static final int TRANSFER_CHUNK_SIZE = 65536;
    
    
    /**
	 * set the sendtimout 
	 * @param sendTimeoutMillis  the send timeout
	 */
	public abstract void setSendTimeoutMillis(long sendTimeoutMillis);
    

    /**
     * {@inheritDoc}
     */
    public abstract void flush() throws IOException;    
   
	
	
	
	
	/**
	 * {@inheritDoc}
	 */
	public abstract void close() throws IOException;
	
	
	/**
     * closes this connection by swallowing io exceptions
     */
    public abstract void closeQuitly();

    
    /**
     * writes a buffer array 
     * 
     * @param buffers                  the buffers to write
     * @param writeCompletionHandler   the completion handler
     * @throws IOException if an exception occurs 
     */
    public abstract void write(ByteBuffer[] buffers, IWriteCompletionHandler writeCompletionHandler) throws IOException;    
    
    
    /**
     * {@inheritDoc}
     */
    public abstract int write(ByteBuffer buffer) throws IOException, BufferOverflowException;
    

    /**
     * {@inheritDoc}
     */
    public abstract long write(ByteBuffer[] buffers) throws IOException, BufferOverflowException;
    
    
    /**
     * {@inheritDoc}
     */
    public abstract long transferFrom(ReadableByteChannel source, int chunkSize) throws IOException, BufferOverflowException; 



    /**
     * {@inheritDoc}
     */
    public abstract long transferFrom(FileChannel fileChannel) throws IOException, BufferOverflowException;	

    
    /**
	 * transfer the available data from the data source 
	 * 
	 * @param source   the data source 
	 * @return the transfered size 
	 * 
	 * @throws IOException if an exception occurs
	 */
	public abstract long transferFrom(NonBlockingBodyDataSource source) throws IOException;

	
	/**
	 * transfer the available data from the data source 
	 * 
	 * @param source   the data source 
	 * @param length   the length to transfer
	 * @return the transfered size 
	 * 
	 * @throws IOException if an exception occurs
	 */
	public abstract long transferFrom(NonBlockingBodyDataSource source, int length) throws IOException;
	
	
	/**
	 * transfer all data from the data source 
	 * 
	 * @param source   the data source 
	 * @return the transfered size 
	 * 
	 * @throws IOException if an exception occurs
	 */
	public abstract long transferFrom(BlockingBodyDataSource source) throws IOException;
	
	
	/**
	 * transfer data from the data source 
	 * 
	 * @param source   the data source 
	 * @param length   the length to transfer 
	 * @return the transfered size 
	 * 
	 * @throws IOException if an exception occurs
	 */
	public abstract long transferFrom(BlockingBodyDataSource source, int length) throws IOException;
	

    /**
     * transfer all data from the data source 
     * 
     * @param source   the data source 
     * @return the transfered size 
     * 
     * @throws IOException if an exception occurs
     */
    public final long transferFrom(InputStream is) throws IOException {
        return transferFrom(Channels.newChannel(is));
    }


    

    /**
     * {@inheritDoc}
     */
    public final long transferFrom(ReadableByteChannel source) throws IOException, BufferOverflowException {
        return transferFrom(source, TRANSFER_CHUNK_SIZE);
    }

    


	/**
	 * {@inheritDoc}
	 */
	public final int write(byte b) throws IOException, BufferOverflowException {
	    return write(new byte[] {b});
	}


	/**
	 * {@inheritDoc}
	 */
	public final int write(byte... bytes) throws IOException, BufferOverflowException {
	    return write(ByteBuffer.wrap(bytes));
	}


	/**
	 * {@inheritDoc}
	 */
	public final int write(byte[] bytes, int offset, int length) throws IOException, BufferOverflowException {
	    return write(DataConverter.toByteBuffer(bytes, offset, length));
	}


	

	/**
	 * {@inheritDoc}
	 */
	public final long write(ByteBuffer[] srcs, int offset, int length) throws IOException {
		return write(DataConverter.toByteBuffers(srcs, offset, length));
	}


	/**
	 * {@inheritDoc}
	 */
	public final long write(List<ByteBuffer> buffers) throws IOException, BufferOverflowException {
		return write(buffers.toArray(new ByteBuffer[buffers.size()]));
	}


	/**
	 * {@inheritDoc}
	 */
	public final int write(int i) throws IOException, BufferOverflowException {
		return write(DataConverter.toByteBuffer(i));
	}


	/**
	 * {@inheritDoc}
	 */
	public final int write(short s) throws IOException, BufferOverflowException {
	    return write(DataConverter.toByteBuffer(s));
	}


	/**
	 * {@inheritDoc}
	 */
	public final int write(long l) throws IOException, BufferOverflowException {
        return write(DataConverter.toByteBuffer(l));
	}


	/**
	 * {@inheritDoc}
	 */
	public final int write(double d) throws IOException, BufferOverflowException {
        return write(DataConverter.toByteBuffer(d));
	}


	/**
	 * {@inheritDoc}
	 */
	public final int write(String message) throws IOException, BufferOverflowException {
        return write(DataConverter.toByteBuffer(message, getEncoding()));
	}

 

	/**
	 * sets the default encoding 
	 * 
	 * @param defaultEncoding  the default encoding 
	 */
	public abstract void setEncoding(String defaultEncoding);


	/**
	 * gets the default encoding 
	 * 
	 * @return  the default encoding
	 */
	public abstract String getEncoding();


	/**
	 * see {@link IConnection#setFlushmode(FlushMode)} 
	 */
	public abstract void setFlushmode(FlushMode flushMode);


	/**
	 * see {@link IConnection#getFlushmode()}
	 */
	public abstract FlushMode getFlushmode();

	
	/**
	 * set autoflush. If autoflush is activated, each write call
	 * will cause a flush. <br><br>
	 *
	 * @param autoflush true if autoflush should be activated
	 */
	public abstract void setAutoflush(boolean autoflush);

	
	
	/**
	 * get autoflush
	 * 
	 * @return true, if autoflush is activated
	 */
	public abstract boolean isAutoflush();
  
	
	
	/**
	 * Marks the write position in the connection. 
	 */
	public abstract void markWritePosition();


	/**
	 * Resets to the marked write position. If the connection has been marked,
	 * then attempt to reposition it at the mark.
	 *
	 * @return true, if reset was successful
	 */
	public abstract boolean resetToWriteMark();


	
	/**
	 * remove the write mark
	 */
	public abstract void removeWriteMark();

	
	/**
	 * Attaches the given object to this connection
	 *
	 * @param obj The object to be attached; may be null
	 * @return The previously-attached object, if any, otherwise null
	 */
	public abstract void setAttachment(Object obj);


	/**
	 * Retrieves the current attachment.
	 *
	 * @return The object currently attached to this key, or null if there is no attachment
	 */
	public abstract Object getAttachment();

	
	
	/**
	 * returns true if the data sink is open
	 * 
	 * @return true if the data sink is open
	 */
	public abstract boolean isOpen();
	

	
    public abstract String getId();
    
    
    /**
     * destroys this data sink  
     */
    public abstract void destroy();

	
	/**
     * add a destroy listener
     * @param destroyListener the destroy listener to add 
     */
    public abstract void addDestroyListener(IBodyDestroyListener destroyListener);
    
    
    /**
     * returns the size of pending write data 
     *   
     * @return the size of pending write data
     */
    abstract int getPendingWriteDataSize();
    
        
    abstract int getSizeWritten();
    
    abstract boolean isNetworkendpoint();
    
    abstract void addCloseListener(IBodyCloseListener closeListener);

}
	    
